// DSKExtract - Command line CoCo DSK extraction utility
// Version 1.0.1 (May 22, 2006)
//
// Freeware, copyright 2006 by Patrick Benny (chipple.net)
// http://www.chipple.net/coco/DSKExtract/

// DESCRIPTION
// ~~~~~~~~~~~
// This utility extracts the contents of one or more regular 35-track CoCo
// virtual disk DSK file(s) into new sub-folders of the same name.

// SYNTAX
// ~~~~~~
//  DSKExtract [dsk_file1.DSK .. [dsk_fileN.DSK]]
//  e.g. DSKExtract mydisk.dsk
//
// In Windows, one or more DSK file(s) can also be simply drag-and-dropped
// onto this program.
//
// Running the program without any arguments will allow extracting 
// the contents of a DSK by inputing its filename.

// SOME LIMITATIONS
// ~~~~~~~~~~~~~~~~
// - Doesn't read disks of formats other than the following:
//   1 side, 35 track, 18 sectors/trk, 256 bytes/sct (JVC format with no header)
// - Disks without a standard directory and GAT will give unexpected results.

// VERSION HISTORY
// ~~~~~~~~~~~~~~~
// 1.0.1  2006/5/22  Now allows filenames containing any character. Characters
//                   other than A-Z a-z 0-9 -#$%&'()@ [space] are converted to _.
// 1.0    2006/4/20  First public release


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <direct.h>

#define C_iFilenameMaxLen     1024     // Maximum filename length

#define C_iDiskSize           0x27600  // Total disk size
#define C_iDiskNbGranules     68       // Number of granules
#define C_iDiskGATTrackPos    0x13200  // GAT track offset
#define C_iDiskGATTrackGATPos 0x100    // GAT offset within GAT track
#define C_iDiskGATTrackDirPos 0x200    // Dir offset within GAT track


// This block for implementing getch() for both Win and Linux
#ifdef WIN32
#include <conio.h>
#else
#include <termios.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>

// Function getch
//          ~~~~~
// Linux (possibly POSIX) replacement for Windows' "conio.h"
//
int getch (void)
{
	/* Ripped from                                             *
	 * <http://cboard.cprogramming.com/showthread.php?t=27714> */
	struct termios oldt, newt;
	int ch;

	tcgetattr (STDIN_FILENO, &oldt);
	newt = oldt;
	newt.c_lflag &= ~(ICANON | ECHO);
	tcsetattr (STDIN_FILENO, TCSANOW, &newt);
	ch = getchar ();
	tcsetattr (STDIN_FILENO, TCSANOW, &oldt);

	return ch;
}
#endif


// Function strtoupper
//          ~~~~~~~~~~
// Converts a string to uppercase.
//
// Parameters:
//    psz  String that will be converted to uppercase
//
// Returns a pointer to converted string (same as input) for convenience
//
char * strtoupper (char *psz) {
    int iLen = strlen(psz);  // Length of string
    int iPos;                // Counter for position
    for (iPos = 0; iPos < iLen; iPos++)
        psz[iPos] = toupper(psz[iPos]);
    return psz;
}

// Function AddFile
//          ~~~~~~~
// Add a file to the disk
//
// Parameters:
//    fIn         Input DSK file handle
//    pszOutPath  Output path string
//
// Returns 0 if ok
//      or 1 if error encountered (input file can't be open, disk full)
//
char ExtractFiles(FILE *fIn,
                  char *pszOutPath) {

    // disk-related data
    unsigned char szDisk[C_iDiskSize];
    unsigned char szGATTrack[0x900*2];

    int iFileNo;
    unsigned char *pszDirFile;
    unsigned char szInFilename[12+1];
    unsigned char szOutFilename[12+1];
    unsigned char szOutPathFile[C_iFilenameMaxLen+1];
    int iFilenameLen;                 // Filename length
    unsigned char cCurGranule;
    unsigned char cNextGranule;

    FILE *fOut;                       // Output file handle
    long iPos;                        // Counter for loop
    int iAnswer;                      // User's answer to getch() question

    unsigned char cTemp;              // One character
    int iRet;                         // Return from fwrite


    // Load disk
    fread(szDisk, sizeof(unsigned char), C_iDiskSize, fIn);

    // Copy GAT track
    memcpy(szGATTrack,&szDisk[C_iDiskGATTrackPos],sizeof(szGATTrack));

    // Move 0x14400+ over GAT track to ease granule access
    memmove(
        &szDisk[C_iDiskGATTrackPos],
        &szDisk[C_iDiskGATTrackPos+0x900*2],
        sizeof(szDisk)-C_iDiskGATTrackPos-0x900*2);

    for (iFileNo = 0; iFileNo < 80; iFileNo++) {
        pszDirFile = &szGATTrack[C_iDiskGATTrackDirPos+iFileNo*0x20];
        cTemp = pszDirFile[0];
        cCurGranule = pszDirFile[0x0d];
        if (  cTemp != 0x00
           && cTemp != 0xff
           && cCurGranule < C_iDiskNbGranules) {

            // Parse filename
            iPos = iFilenameLen = 0;
            while (iPos < 8) {
                szInFilename[iFilenameLen++] = cTemp;
                cTemp = pszDirFile[++iPos];
            }
            while (szInFilename[--iPos] == 0x20)
                szInFilename[iPos] = 0x00;
            iFilenameLen = iPos+1;

            // Parse extension
            szInFilename[iFilenameLen++] = '.';
            iPos = 8;
            while (iPos < 11) {
                szInFilename[iFilenameLen++] = cTemp;
                cTemp = pszDirFile[++iPos];
            }
            szInFilename[iFilenameLen] = 0;
            while (szInFilename[--iPos] == 0x20)
                szInFilename[iPos] = 0x00;
            iFilenameLen = iPos + 1;

            // Replace unallowed characters with "_"
            strcpy(szOutFilename,szInFilename);
            iPos = 0;
            for (iPos = 0; iPos < iFilenameLen; iPos++)
                if (strchr("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-#$%&'()@ .",szOutFilename[iPos]) == NULL)
                    szOutFilename[iPos] = '_';

            if (strcmp(szOutFilename,szInFilename) != 0)
                printf("Extracting file %s as %s\n",szInFilename,szOutFilename);
            else
                printf("Extracting file %s\n",szOutFilename);

            strcpy(szOutPathFile,pszOutPath);
            strcat(szOutPathFile,szOutFilename);

            // Check whether output file exists
            if (iAnswer != 'A' && (fOut = fopen(szOutPathFile,"r")) != NULL) {
                fclose(fOut);
                printf("File already exists, overwrite? Y/[N]/All ");
                while ((iAnswer = toupper(getch())) != 'Y'
                    && iAnswer != 'N' 
                    && iAnswer != 'A'
                    && iAnswer != 13    // Enter
                    && iAnswer != 32    // Space
                    && iAnswer != 3     // CTRL+C
                    && iAnswer != -1);  // Break?
                if (iAnswer == 3 || iAnswer == -1)
                    return 1;
                if (iAnswer != 'Y' && iAnswer != 'A') {
                    printf("\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8N        \n");
                    continue;
                }
                printf("\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8%c        \n",iAnswer);
            }

            // Open output file
            if ((fOut = fopen(szOutPathFile,"wb")) == NULL) {
                perror("Error opening output file, aborting");
                return 1;
            }

            // Write all granules completely, except last one
            while ((cNextGranule = szGATTrack[C_iDiskGATTrackGATPos+cCurGranule]) < C_iDiskNbGranules) {
                iRet = fwrite(
                          &szDisk[0x900*cCurGranule],
                          sizeof(unsigned char),
                          0x900,
                          fOut);
                cCurGranule = cNextGranule;
            }

            // Write last granule
            iRet = fwrite(
                      &szDisk[0x900*cCurGranule],
                      sizeof(unsigned char),
                      (cNextGranule-0xc1+pszDirFile[0x0e])*0x100+pszDirFile[0x0f],
                      fOut);

            // Close output file
            fclose(fOut);
        }
    }

    return 0;
}

// Main function
int main(int argc, char **argv) {
    FILE *infile = NULL;      // Input DSK file handle
    int iArg;                 // Counter to loop through arguments
    unsigned char szInFile[C_iFilenameMaxLen+1];   // Input DSK filename
    char cFirstTry = 1;       // 1 if first try inputting filename, otherwise 0
    unsigned char *pszExt;    // Extension pos.
    unsigned char szOutPath[C_iFilenameMaxLen+1];  // Output path
    char cRet = 0;            // Return code of ExtractFiles()

#ifdef WIN32
    struct _stat statFile;
#else
    struct stat statFile;
#endif


    // Display header
    printf("DSKExtract - Command line CoCo DSK extraction utility\n");
    printf("Version 1.0.1 (May 22, 2006)\n");
    printf("Freeware, copyright 2006 by Patrick Benny\n");
    printf("http://www.chipple.net/coco/DSKExtract/\n\n");

    printf("This utility extracts the contents of regular 35-track CoCo virtual disk\nDSK files into new sub-folders of the same name.\n\n");

    if (argc <= 1) {
        // Display instructions
        printf("Syntax:\n");
        printf("  DSKExtract [dsk_file.DSK .. [dsk_fileN.DSK]]\n");
        printf("  e.g. DSKExtract mydisk.dsk\n");
        printf("In Windows, DSK files can also be simply drag-and-dropped onto this program.\n\n");

        printf("To extract files from a DSK, enter DSK filename. (Ctrl-C to quit)\n");

        // Ask for filename until a valid filename is input
        do {
            // Display error message if not first try
            if (cFirstTry == 0)
                perror("Error opening input file");
            else
                cFirstTry = 0;
            // Get user input
            printf("DSK file: ");
            fgets(szInFile, C_iFilenameMaxLen, stdin);
            // Remove trailing new line
            szInFile[strlen(szInFile)-1] = '\0';
        }
        while (szInFile[0] != '\0' && (infile = fopen(szInFile, "rb")) == NULL);
        if (infile != NULL)
            fclose(infile);

        // If blank string input, abort
        if (szInFile[0] == '\0') {
            printf("Aborted.");
            exit(1);
        }
        strcpy(szInFile,szInFile);
        printf("\n");
    }

    for (iArg = 1; cRet == 0 && iArg < (argc<=1?2:argc); iArg++) {
        if (argc > 1)
            strcpy(szInFile,argv[iArg]);

        printf("Input DSK file: %s\n", szInFile);

        strcpy(szOutPath,szInFile);
        if ((pszExt = strrchr(szOutPath,'.')) != NULL) {
            strtoupper(pszExt);
            if (strcmp(pszExt,".DSK") != 0) {
                printf("Input file must have a DSK extension, skipping.\n");
                continue;
            }
        }

        pszExt[0] = '\\';
        pszExt[1] = '\0';
        mkdir(szOutPath);

        // Open input file
        if ((infile = fopen(szInFile, "rb")) == NULL) {
            perror("Error opening input file, skipping");
            continue;
        }

        // Check file size
#ifdef WIN32
        _stat(szInFile, &statFile);
#else
        stat(szInFile, &statFile);
#endif
        if (statFile.st_size != 35*18*256) {
            printf("Error: Not a 35-track 18-sector disk (len: %l), skipping.\n");
            fclose(infile);
            continue;
        }

        // Extract files from DSK
        printf("Output folder: %s\n", szOutPath);
        cRet = ExtractFiles(infile,szOutPath);

        // Close input file
        fclose(infile);

        printf("\n");
    }

    // Check if disk extraction successful
    if (cRet == 0)
        printf("Done.\n");
    else
        printf("Aborted.\n");

    exit(cRet);
}
